from typing import Any, Dict, Type, Union

import pytest
from gamma.dispatch import DispatchError, dispatch
from gamma.dispatch.dispatchsystem import OverwriteWarning, Sig, methods_matching
from gamma.dispatch.typesystem import Sig, SigSet, Vararg, signatures_from


class SuperFoo:
    pass


class Foo(SuperFoo):
    pass


class SuperBar:
    pass


class Bar(SuperBar):
    pass


def test_simple_dispatch():
    @dispatch
    def temp(a):
        return "fallback"

    @dispatch
    def temp(a: int):
        return "int"

    @dispatch
    def temp(a: float):
        return "float"

    @dispatch
    def temp(a: float, b: str):
        return "float-str"

    assert temp("a") == "fallback"
    assert temp(1) == "int"
    assert temp(1.0) == "float"
    assert temp(1.0, "foo") == "float-str"

    with pytest.raises(DispatchError, match="no method found"):
        temp(1.0, [])


def test_wrong_kwarg_call():
    @dispatch
    def temp(*, c=1):
        return "noarg"

    @dispatch
    def temp(a):
        return "fallback"

    @dispatch
    def temp(a: int):
        return "int"

    assert temp("a") == "fallback"
    assert temp(1) == "int"

    with pytest.raises(DispatchError, match="[Rr]eserved"):
        temp(a=1)

    with pytest.raises(TypeError, match="unexpected keyword"):
        temp(b=1)

    assert temp(c=1) == "noarg"


def test_warn_overwrite():
    @dispatch
    def temp(a):
        return "1 arg"

    with pytest.warns(OverwriteWarning):
        # common mistake to accidentally overwrite a function call
        @dispatch
        def temp(c=1):
            return "overwrites"

    @dispatch(overwrite=True)
    def temp(c=1):
        return "overwrites"

    assert temp(1) == "overwrites"


def test_full_order():
    @dispatch
    def temp(a, b: int):
        return "1"

    @dispatch
    def temp(a: int, b):
        return "2"

    assert temp([], 1) == "1"
    assert temp(1, []) == "2"

    # this is ambiguous
    with pytest.raises(DispatchError):
        temp(1, 1)

    # solve ambiguity
    @dispatch
    def temp(a: int, b: int):
        return "3"

    assert temp(1, 1) == "3"


def test_varargs():
    @dispatch
    def temp(a: int, b: int, *args):
        return f"{a}, {b}"

    assert temp(1, 2) == "1, 2"
    assert temp(1, 2, 3) == "1, 2"
    assert temp(1, 2, 3, 4) == "1, 2"

    with pytest.raises(DispatchError):
        temp(1)

    @dispatch
    def temp(a: int, b: int, c: str, *args):
        return "new"

    assert temp(1, 2, 3) == "1, 2"
    assert temp(1, 2, "3") == "new"


def test_defaults():
    @dispatch
    def temp(a: int, b: int = 1):
        return f"{a}, {b}"

    assert temp(1, 1) == "1, 1"
    assert temp(1, 2) == "1, 2"
    assert temp(1) == "1, 1"


def test_specialization():
    class AbstractFoo:
        pass

    class Foo(AbstractFoo):
        pass

    @dispatch
    def temp(a: AbstractFoo, b: bool):
        return "2 args"

    @dispatch
    def temp(a: Foo):
        return "1 arg"

    assert temp(Foo(), False) == "2 args"
    assert temp(Foo()) == "1 arg"


def test_dispatch_exception():
    @dispatch
    def temp(x: int):  # noqa
        return "int"

    @dispatch
    def temp(x: float):  # noqa
        return "float"

    with pytest.raises(TypeError, match="foo"):
        # invalid number of args, check source file is part of the exception args
        temp(1, foo=1)


def test_name_shadowing():
    # an object with the same name appearing previously in the same namespace
    temp = 123  # noqa

    # a multi-dipatch method shadowing that name
    @dispatch
    def temp(x: int):  # noqa
        return "int"

    @dispatch
    def temp(x: float):
        return "float"

    assert isinstance(temp, dispatch)
    assert temp(0) == "int"
    assert temp(0.0) == "float"


def test_union_sig():
    def temp(x: Union[int, float]):  # noqa
        return "int-float"

    for sig in signatures_from(temp):
        assert "int" in repr(sig)
        assert "float" in repr(sig)


def test_union_call():
    @dispatch
    def temp(x: Union[int, float]):  # noqa
        return "int-float"

    @dispatch
    def temp(x):  # noqa
        return "fallback"

    assert temp(1) == "int-float"
    assert temp(1.0) == "int-float"
    assert temp([]) == "fallback"


def test_inherit_none():
    class Base:
        pass

    class Foo(Base):
        pass

    @dispatch
    def temp(x: Base, y=None):
        return "foo"

    assert temp(Foo()) == "foo"
    assert temp(Foo(), None) == "foo"
    assert temp(Foo(), 1) == "foo"


def test_typing_alias():
    @dispatch
    def temp(x: Dict):
        return "dict"

    @dispatch
    def temp(x):
        return "fallback"

    assert temp(dict(a=1)) == "dict"
    assert temp(1) == "fallback"


def test_cache_reset():
    @dispatch
    def temp(x):
        return "obj"

    assert temp(1) == "obj"

    @dispatch
    def temp(x: int):
        return "int"

    assert temp(1) == "int"


def test_pending():
    # forward reference to types won't work for local classes
    assert global_temp(GlobalFoo()) == "foo"


@dispatch
def global_temp(x: "GlobalFoo"):
    return "foo"


class GlobalFoo:
    pass


def test_manual_setitem():
    @dispatch
    def temp(x: int):
        return "int"

    def alt(*args):
        return "alt"

    temp[str] = alt
    temp[float, float] = alt

    assert temp(1) == "int"
    assert temp("1") == "alt"
    assert temp(1.0, 1.0) == "alt"


def test_specialize1():
    sig1 = Sig(object, object)
    sig2 = Sig(SuperFoo, object)
    sig3 = Sig(Foo, object)
    sig4 = Sig(Foo, SuperFoo)
    sigN = Sig(int, int)
    table = SigSet([sig1, sig2, sig3, sig4, sigN])

    call = Sig(Foo, Foo)
    match = methods_matching(call, table)
    assert match[0] == sig4
    assert sigN not in match
    assert len(match) == 1

    # add an ambiguous match
    # (:Foo, :Foo) matches (:Foo, :SuperFoo) and (:SuperFoo, :Foo)
    # but one is not a strict subtype of the other
    sig5 = Sig(SuperFoo, Foo)
    table.add(sig5)
    match = methods_matching(call, table)
    assert len(match) == 2

    # add a more specific sig
    sig6 = Sig(Foo, Foo)
    table.add(sig6)
    match = methods_matching(call, table)
    assert len(match) == 1


def test_specialize2():
    sig1 = Sig(Foo)
    sig2 = Sig(SuperFoo, bool)
    table = SigSet([sig1, sig2])

    call = Sig(Foo, bool)
    got = methods_matching(call, table)
    assert len(got) == 1
    assert got[0] == sig2

    call = Sig(Foo)
    got = methods_matching(call, table)
    assert len(got) == 1
    assert got[0] == sig1


def test_varargs_dispatch():
    target = Sig(int, int, Vararg)
    call = Sig(int, int, int)
    match = methods_matching(call, [target])
    assert match

    target = Sig(int, int, Vararg)
    call = Sig(int, int)
    match = methods_matching(call, [target])
    assert match


def test_caching():
    @dispatch
    def temp(a):
        return 1

    assert temp("a") == 1

    @dispatch
    def temp(a: str):
        return 2

    assert temp("a") == 2
    assert temp(True) == 1


def test_dispatch_on_types():
    @dispatch
    def temp(a):
        return "fallback"

    @dispatch
    def temp(a: type):
        return "anytype"

    @dispatch
    def temp(a: Type[Bar]):
        return "bar"

    @dispatch
    def temp(a: Type[Foo]):
        return "foo"

    assert temp(1) == "fallback"
    assert temp(Foo) == "foo"
    assert temp(Bar) == "bar"
    assert temp(str) == "anytype"


def test_dispatch_on_any():
    @dispatch
    def temp1(a: Any):
        return "fallback"

    @dispatch
    def temp1(a: int):
        return 1

    #

    @dispatch
    def temp2(a: Any, b: int):
        return "fallback"

    @dispatch
    def temp2(a: int, b: int):
        return 2

    assert temp1(0) == 1
    assert temp1(...) == "fallback"
    assert temp2(0, 0) == 2
    assert temp2(..., 0) == "fallback"



from typing import List

from gamma.config import RootConfig, to_dict
from gamma.config.pydantic import ConfigStruct


class Dataset(ConfigStruct):
    """Base class with common properties."""

    path: str
    compression: str


class CsvDataset(Dataset):
    kind = "csv"
    separator: str


class ParquetDataset(Dataset):
    kind = "parquet"
    columns: List[str]


def test_auto():
    src = """
datasets:
  foo:
    kind: csv
    path: data/foo.csv.gz
    compression: gzip
    separator: ";"

  bar:
    kind: parquet
    path: data/bar.parquet
    compression: snappy
    columns: [col_x, col_y]
    """

    config = RootConfig("dummy", src)
    get_config = lambda: config  # noqa

    def get_dataset(name):
        entry = get_config()["datasets"][name]
        obj = Dataset.parse_obj(to_dict(entry))
        return obj

    foo = get_dataset("foo")
    assert isinstance(foo, CsvDataset)
    assert foo.separator == ";"
    assert isinstance(get_dataset("bar"), ParquetDataset)



import io

from gamma.config import RootConfig, to_dict, to_yaml
from gamma.config.tags import Tag
from gamma.dispatch import dispatch
from ruamel.yaml import YAML
from ruamel.yaml.nodes import MappingNode


def test_to_dict_basic():
    # test simple case
    src = dict(a=1, b=2)
    cfg = RootConfig("dummy", src)
    d = to_dict(cfg)
    assert d == src
    assert type(d) == dict

    # test nested dicts
    src = dict(a=dict(b=1))
    cfg = RootConfig("dummy", src)
    d = to_dict(cfg)
    assert d == src
    assert type(d) == dict
    assert type(d["a"]) == dict

    # test nested dicts within lists
    src = dict(a=[dict(b=1), dict(c=2), dict(d=[dict(e=1)])])
    cfg = RootConfig("dummy", src)
    d = to_dict(cfg)
    assert d == src
    assert type(d) == dict
    assert type(d["a"]) == list
    assert type(d["a"][0]) == dict
    assert type(d["a"][1]) == dict
    assert type(d["a"][2]) == dict
    assert type(d["a"][2]["d"]) == list
    assert type(d["a"][2]["d"][0]) == dict
    assert type(d["a"][2]["d"][0]["e"]) == int


def test_custom_dict_tag():
    src = """
    foo: !customdict
        a: 1
    """

    from gamma.config.render import render_node

    CustomTag = Tag["!customdict"]
    try:

        @dispatch
        def render_node(node: MappingNode, tag: CustomTag, **ctx):
            # force render to dict even if tagged
            return to_dict(node)

        cfg = RootConfig("dummy", src)
        d = cfg["foo"]
        assert type(d) == dict
        assert d["a"] == 1
    finally:
        del render_node[MappingNode, CustomTag]


def test_to_yaml_basic():
    # test simple case
    src = dict(a=1, b=2)
    cfg = RootConfig("dummy", src)
    content = to_yaml(cfg)
    d = dict(YAML().load(io.StringIO(content)))
    assert d == src



import pytest
from gamma.dispatch import ParametricMeta, Val, dispatch, parametric
from gamma.dispatch.dispatchsystem import Sig, issubtype


class Global(metaclass=ParametricMeta):
    pass


@parametric
class Global2:
    pass


def test_global():
    assert Global.__module__ == __name__
    assert Global["a"].__module__ == __name__

    assert Global2.__module__ == __name__
    assert Global2["a"].__module__ == __name__


def test_parametric_create():
    class Tag(metaclass=ParametricMeta):
        pass

    RefTag1 = Tag["ref"]
    RefTag2 = Tag.of("ref")
    assert RefTag1 is Tag["ref"] is RefTag2

    NamedTag1 = Tag.of("named", name="NamedTag")
    NamedTag2 = Tag["named"]
    NamedTag3 = Tag.of("named", name="NamedTag")
    assert NamedTag1 is NamedTag2 is NamedTag3

    with pytest.raises(NameError):
        Tag.of("named", name="OtherName")

    assert issubclass(RefTag1, Tag)
    assert RefTag1.__values__ == ("ref",)
    assert RefTag1.__module__ == __name__


def test_parametric_deco():
    from typing import NamedTuple

    @parametric
    class Tag:
        pass

    RefTag1 = Tag["ref"]
    RefTag2 = Tag.of("ref")
    assert RefTag1 is Tag["ref"] is RefTag2

    # test interaction with namedtuple
    @parametric
    class FooType(NamedTuple):
        foo: str

    foo1 = FooType["A"]("arg1")
    foo2 = FooType["A"]("arg2")
    foo1a = FooType["A"]("arg1")
    assert foo1.foo == "arg1" and foo2.foo == "arg2"
    assert foo1 == foo1a


def test_parametric_dispatch():
    # create value type hierarchy
    @parametric
    class Tag:
        pass

    RefTag = Tag["ref"]
    FooTag = Tag.of("foo")
    assert RefTag is not FooTag

    # create functions
    @dispatch
    def fun(a):
        return "fallback"

    @dispatch
    def fun(a: Tag):
        return "base"

    @dispatch
    def fun(a: RefTag):
        return "ref"

    @dispatch
    def fun(a: Tag["foo"]):  # noqa: F821
        return "foo"

    assert fun(1) == "fallback"
    assert fun(Tag["new"]()) == "base"
    assert fun(RefTag()) == "ref"
    assert fun(FooTag()) == "foo"

    @dispatch
    def bar(a: RefTag):
        return "ok"

    assert bar(RefTag()) == "ok"
    assert bar(RefTag) == "ok"


def test_also_of():
    from gamma.dispatch import parametric

    # create value type hierarchy
    @parametric
    class Tag:
        pass

    ATag = Tag["A"]
    BTag = ATag.also_of("B")
    assert ATag is BTag
    assert Tag["A"] is Tag["B"]


def test_union():
    from gamma.dispatch import parametric, dispatch
    from typing import Union

    @parametric
    class Tag:
        pass

    ATag = Tag["A"]
    BTag = Tag["B"]
    AorB = Union[ATag, BTag]

    @dispatch
    def fun(a):
        return "fail"

    @dispatch
    def fun(a: AorB):
        return "ok"

    assert fun(ATag()) == "ok"
    assert fun(BTag()) == "ok"
    assert fun(Tag["C"]()) == "fail"


def test_type_system():
    @parametric
    class Tag:
        pass

    ATag = Tag["A"]
    BTag = Tag["B"]

    assert issubtype(ATag, Tag)
    assert issubtype(BTag, Tag)

    a = Sig(ATag)
    b = Sig(BTag)
    base = Sig(Tag)

    assert not issubtype(a, b)
    assert issubtype(a, base)


def test_dispatch_valuetype():
    ATag = Val["A"]

    @dispatch
    def fun(a):
        # should return an instance regardless if a type was passed
        assert isinstance(a, Val)
        assert isinstance(a, ATag)
        return "ok"

    # should dispatch parametric types on instance or on the type itself
    assert fun(ATag) == "ok"
    assert fun(ATag()) == "ok"



from gamma.config.confignode import RootConfig, get_keys, push_entry
from gamma.config.load import load_node
from gamma.config.merge import merge_nodes
from gamma.config.render import render_node


def test_root_config_merge():
    root = RootConfig()

    a = "{foo: bar}"
    b = "{foo: baz, ping: pong}"
    push_entry(root, "a1", a)
    push_entry(root, "b1", b)
    assert root["foo"] == "baz"
    assert root["ping"] == "pong"

    keys = set([render_node(k) for k in get_keys(root)])
    assert keys == set(["foo", "ping"])


def test_merge():

    # test simple
    def _test(a, b, expect):
        left = "target", load_node(a)
        right = "patch", load_node(b)
        _, node = merge_nodes(*left, *right)
        got = render_node(node)
        assert got == expect

    a = "foo: bar"
    b = "foo: baz"
    expect = {"foo": "baz"}
    _test(a, b, expect)

    # test dict merging
    a = "foo: {a: 1, b: 2}"
    b = "foo: {b: 20, c: 30}"
    expect = {"foo": {"a": 1, "b": 20, "c": 30}}
    _test(a, b, expect)

    # test list merging
    a = "foo: [1, 2]"
    b = "foo: [2, 3]"
    expect = {"foo": [1, 2, 3]}
    _test(a, b, expect)

    # tes list of objects merging
    a = "foo: [{a: 1, b: 1}, {c: 1, d: 1}]"
    b = "foo: [{a: 2, b: 1}, {c: 1, d: 1}]"
    expect = {"foo": [{"a": 1, "b": 1}, {"c": 1, "d": 1}, {"a": 2, "b": 1}]}
    _test(a, b, expect)

    # test type replacement (1)
    a = "foo: [1, 2]"
    b = "foo: {b: 20, c: 30}"
    expect = {"foo": {"b": 20, "c": 30}}
    _test(a, b, expect)


def test_hints():

    # assert replace hint on inline lists
    target = load_node({"foo": [1, 2]})
    patch = load_node(
        """
        foo: [2, 3] # @hint: merge_replace
        """
    )
    _, node = merge_nodes(target, patch)
    assert render_node(node) == {"foo": [2, 3]}

    # assert replace hint on expanded lists
    target = load_node({"foo": [1, 2]})
    patch = load_node(
        """
        foo: # @hint: merge_replace
          # this comment should be ok
          - 2
          - 3
        """
    )
    _, node = merge_nodes(target, patch)
    assert render_node(node) == {"foo": [2, 3]}

    # test dict merging inline map
    target = load_node({"foo": {"a": 1, "b": 2}})
    patch = load_node(
        """
        foo: {"b": 20, "c": 30} # @hint: merge_replace
        """
    )
    _, node = merge_nodes(target, patch)
    assert render_node(node) == {"foo": {"b": 20, "c": 30}}

    # test dict merging expanded map
    target = load_node({"foo": {"a": 1, "b": 2}})
    patch = load_node(
        """
        foo:  # @hint: merge_replace
          b: 20
          c: 30
        """
    )
    _, node = merge_nodes(target, patch)
    assert render_node(node) == {"foo": {"b": 20, "c": 30}}

    # test some hint syntax variants - 1
    target = load_node({"foo": {"a": 1, "b": 2}})
    patch = load_node(
        """
        foo: {"b": 20, "c": 30} #@hint: merge_replace
        """
    )
    _, node = merge_nodes(target, patch)
    assert render_node(node) == {"foo": {"b": 20, "c": 30}}

    # test some hint syntax variants - 2
    target = load_node({"foo": {"a": 1, "b": 2}})
    patch = load_node(
        """
        foo: {"b": 20, "c": 30} #@hint:   merge_replace
        """
    )
    _, node = merge_nodes(target, patch)
    assert render_node(node) == {"foo": {"b": 20, "c": 30}}



import os


def test_pickle(monkeypatch):
    from gamma.config import get_config
    import pickle

    c1 = get_config()
    data = pickle.dumps(c1)
    assert data

    c2 = pickle.loads(data)

    assert c2 is not c1

    # check still dynamic
    assert c2["sample_env"]["user"] == os.getenv("USER")
    monkeypatch.setenv("USER", "foo")
    assert c2["sample_env"]["user"] == "foo"


def _sub(c, expected):
    assert c["sample_env"]["user"] == expected


def test_subprocess():
    import multiprocessing
    from gamma.config import get_config

    c = get_config()
    expected = c["sample_env"]["user"]
    p = multiprocessing.Process(target=_sub, args=(c, expected))
    p.start()
    p.join()
    assert p.exitcode == 0



from gamma.config import RootConfig, to_dict
from pydantic import BaseModel


class Dataset(BaseModel):
    path: str
    compression: str


def test_manual():
    src = """
datasets:
  foo:
    path: data/foo.csv.gz
    compression: gzip
  bar:
    path: data/bar.parquet
    compression: snappy
    """

    config = RootConfig("dummy", src)
    get_config = lambda: config  # noqa

    def get_dataset(name):
        entry = get_config()["datasets"][name]
        obj = Dataset(**to_dict(entry))
        return obj

    assert isinstance(get_dataset("foo"), Dataset)
    assert isinstance(get_dataset("bar"), Dataset)



import os

import pytest


def test_load_sample(caplog, monkeypatch):
    import logging

    caplog.set_level(logging.DEBUG)

    from gamma.config import get_config
    from gamma.config.globalconfig import reset_config

    # load default config
    config = get_config()

    # assert meta was loaded
    config["include_folders"]  # should not fail

    # assert default was loaded
    assert config["sample_scalar_1"] == "hello world"
    assert config["sample_list_1"] == [1, 2, 3]

    # assert env was loaded and override default
    assert config["sample_scalar_2"] == "foobar_dev"
    assert config["deep"]["lvl1"]["lvl2"]["lvl3"] == "foobar_dev"

    # assert dot access
    config._dot_access = True
    assert config.sample_list_1 == [1, 2, 3]

    with pytest.raises(AttributeError):
        assert not config.missing

    with pytest.raises(AttributeError):
        assert not config._missing_underscore
    config._dot_access = False

    # reset config and test other env
    reset_config()
    monkeypatch.setenv("ENVIRONMENT", "prod")
    config = get_config()
    assert config["sample_scalar_2"] == "foobar_prod"
    assert config["deep"]["lvl1"]["lvl2"]["lvl3"] == "foobar_prod"
    assert config["deep"]["lvl1"]["lvl2"]["new"] == "new_prod"


def test_env_default(monkeypatch):
    from gamma.config import get_config

    config = get_config()

    # unset USER env to test default
    monkeypatch.delenv("USER", None)

    with pytest.raises(Exception):
        assert config["sample_env"]["user"]
    assert config["sample_env"]["default_1"] == "foo"
    assert config["sample_env"]["default_2"] is None


def test_sample_dump():
    from gamma.config import get_config, to_yaml, to_dict
    from ruamel.yaml import YAML

    # load default config
    config = get_config()
    assert config["sample_scalar_1"] == "hello world"

    # dump resolving tags
    dump = to_yaml(config, True)

    # load in a regular yaml loader
    yaml = YAML(typ="rt")
    loaded = yaml.load(dump)
    assert loaded["sample_scalar_1"] == "hello world"

    # test env
    assert loaded["sample_env"]["user"] == os.environ["USER"]

    # assert secret env was not dumped
    assert not isinstance(loaded["sample_env"]["secret_user"], str)
    assert isinstance(loaded["nested"]["composite"], str)
    assert not isinstance(loaded["nested"]["secret"], str)

    # dump again, with original tags
    dump = to_yaml(config, False)
    yaml = YAML(typ="rt")
    loaded = yaml.load(dump)
    assert not isinstance(loaded["sample_env"]["user"], str)
    assert hasattr(loaded["sample_env"]["user"], "tag")

    # dump partial config
    dump = to_yaml(config["sample_env"], True)
    yaml = YAML(typ="rt")
    loaded = yaml.load(dump)
    assert loaded["user"] == os.environ["USER"]

    # dump to dict
    new_dict = to_dict(config)
    assert type(new_dict) == dict
    assert type(new_dict["sample_env"]) == dict


def test_expression(monkeypatch):
    monkeypatch.setenv("USER", "dummy")
    from gamma.config import get_config
    import os

    config = get_config()

    assert config["sample_expr"]["expr_1"] == 2
    assert config["sample_expr"]["expr_2"] == "user=dummy"
    e3 = config["sample_expr"]["expr_3"]
    assert e3 == os.environ["USER"]



from typing import List

from gamma.config.pydantic import ConfigStruct, URIConfigStruct
from pydantic import BaseModel
from ruamel.yaml import YAML

CONFIG = """
objects:
    foo:
        kind: foo
        value: 100
    bar:
        kind: bar
        custom: 200
    foo_str: 300

lists:
    - kind: foo
      value: 100
    - kind: bar
      custom: 200

myuri: abc:hello/world
myuri2:
    uri: abc:hello/world
"""

yaml = YAML()


class CustomType(ConfigStruct):
    @classmethod
    def parse_scalar(cls, value):
        return Foo(value=value)


class Foo(CustomType):
    kind = "foo"
    value: int


class Bar(CustomType):
    kind = "bar"
    custom: int


class Objects(BaseModel):
    foo: CustomType
    bar: CustomType
    foo_str: CustomType


class MyURI(URIConfigStruct):
    scheme: str = "abc"


class _TestConfig(BaseModel):
    objects: Objects
    lists: List[CustomType]
    myuri: MyURI
    myuri2: MyURI


def test_load_pydantic():
    config = _TestConfig(**yaml.load(CONFIG))
    objs = config.objects
    assert objs
    assert objs.foo.value == 100
    assert objs.bar.custom == 200
    assert objs.foo_str.value == 300

    lists = config.lists
    lists[0].kind == "foo"
    lists[1].kind == "bar"

    assert config.myuri.scheme == "abc"
    assert config.myuri2.scheme == "abc"



import pytest


@pytest.fixture(autouse=True)
def reset_config():
    """Reset config singletons on each test"""
    from gamma.config.globalconfig import reset_config

    yield

    reset_config()


@pytest.fixture(autouse=True)
def autoset_user(monkeypatch):
    monkeypatch.setenv("USER", "dummy")
    yield



from gamma.config import RootConfig


def test_anchors():
    test = """
foo: &foo
  sub1: 10
  sub2: 20

bar:
  <<: *foo
  sub2: 22
  sub3: 33
  subx: *foo

val1: &val1 100

l1:
    - *val1
    - 200
    - 300
"""

    cfg = RootConfig(None, test)

    assert cfg["l1"] == [100, 200, 300]

    assert cfg["bar"]["sub1"] == 10
    assert cfg["bar"]["sub2"] == 22
    assert cfg["bar"]["sub3"] == 33
    assert cfg["bar"]["subx"] == {"sub1": 10, "sub2": 20}

    assert cfg["foo"]["sub1"] == 10
    assert cfg["foo"]["sub2"] == 20



import pytest


def test_as_node():
    from gamma.config.rawnodes import as_node
    from gamma.config import render_node

    # just check reflexivity
    assert render_node(as_node(1)) == 1
    assert render_node(as_node("a")) == "a"
    assert render_node(as_node(1.0)) == 1.0
    assert render_node(as_node(None)) is None
    assert render_node(as_node(False)) is False
    assert render_node(as_node(True)) is True

    a = as_node(as_node(1))
    b = as_node(1)
    assert a.tag == b.tag and a.value == b.value

    class Custom:
        pass

    with pytest.raises(Exception):
        as_node(Custom())



import os
import shutil
import subprocess
from pathlib import Path

import pytest


@pytest.fixture
def prepare_fixture():
    here = Path(".").absolute()
    tmp = Path("./_tmp").absolute()
    tmp.mkdir(exist_ok=True)

    yield {"here": str(here), "_tmp": str(tmp)}

    os.chdir(here)
    # shutil.rmtree(tmp)


def test_find_jupyter(monkeypatch, prepare_fixture):
    from gamma.config.findconfig import get_config_root, FindJupyter
    from gamma.config import findconfig as mod

    expected_root = str(get_config_root())
    assert expected_root is not None

    # change to subfolder, can't find
    os.chdir("_tmp")
    with pytest.raises(Exception):
        get_config_root()

    # use findJupyter specialization
    monkeypatch.setattr(mod, "_isnotebook", lambda: True)
    jupyter_root = str(get_config_root(FindJupyter()))
    assert expected_root == jupyter_root

    os.chdir("..")

    # call using python in subfolder to assert we can't get a config root
    script = Path("_tmp") / "script.py"
    script.write_text(
        """
from gamma.config.findconfig import get_config_root
try:
    print(get_config_root())
except:
    print("None")
"""
    )
    env = os.environ.copy()
    env["PYTHONPATH"] = prepare_fixture["here"]
    cwd = "_tmp"
    python = shutil.which("python")
    cmd = [python, "script.py"]
    cp = subprocess.run(cmd, check=True, capture_output=True, env=env, cwd=cwd)
    out = cp.stdout.decode().strip()
    assert out == "None"

    # call using ipython in subfolder to check we CAN get a config root
    ipython = shutil.which("ipython")
    cmd = [ipython, "script.py"]
    cp = subprocess.run(cmd, check=True, capture_output=True, env=env, cwd=cwd)
    out = cp.stdout.decode().strip()
    assert out.endswith(expected_root)



from timeit import default_timer as timer

from gamma.dispatch import dispatch


class Base:
    def __call__(self, *args, **kwargs):
        return "fallback"


def test_simple_dispatch():
    @dispatch
    def temp(a):
        return "fallback"

    @dispatch
    def temp(a: int):
        return "int"

    base = Base()

    # warm
    base(1)
    temp(1)

    n = 100_000
    start = timer()
    for _ in range(n):
        base(1)
    base_t = timer() - start

    start = timer()
    for _ in range(n):
        temp(1)
    test_t1 = timer() - start

    print(base_t)
    print(test_t1)
    print(test_t1 / base_t)
    assert test_t1 / base_t < 16


def test_kwargs():
    @dispatch
    def temp(a, *, c=1):
        return "fallback"

    @dispatch
    def temp(a: int, *, c=1):
        return "int"

    base = Base()

    # warm
    base(1, c=2)
    temp(1, c=2)

    n = 100_000
    start = timer()
    for _ in range(n):
        base(1, c=2)
    base_t = timer() - start

    start = timer()
    for _ in range(n):
        temp(1, c=2)
    test_t1 = timer() - start

    print(base_t)
    print(test_t1)
    print(test_t1 / base_t)
    assert test_t1 / base_t < 15



from gamma.config import RootConfig
from pydantic import BaseModel


class Dataset(BaseModel):
    path: str
    compression: str


def test_obj():
    from tests.gamma.config.pydantic.test_doc2 import Dataset

    src = """
obj_default_module: tests.gamma.config.pydantic.test_doc2

datasets:
  foo: !obj:Dataset
    path: data/foo.csv.gz
    compression: gzip
  bar: !obj:Dataset
    path: data/bar.parquet
    compression: snappy
    """

    config = RootConfig("dummy", src)
    get_config = lambda: config  # noqa

    foo = get_config()["datasets"]["foo"]
    bar = get_config()["datasets"]["bar"]
    assert isinstance(foo, Dataset)
    assert isinstance(bar, Dataset)



import os
import sys
import tempfile
from pathlib import Path

import pytest
from gamma.config import get_config

CONFIG_STR = """
test: !myenv USER
test2: !env USER
"""


@pytest.fixture
def folder_fixture(monkeypatch):

    with tempfile.TemporaryDirectory() as td:
        base = Path(td)

        # create config/00-meta.yaml
        root = base / "config"
        root.mkdir()
        meta = root / "00-meta.yaml"
        meta.write_text("include_folders: []")

        # create config/10-test.yaml
        test_cfg = root / "10-test.yaml"
        test_cfg.write_text(CONFIG_STR)

        # create foo package
        foo = base / "foo"
        foo.mkdir()

        # copy plugin contents
        plugin = foo / "custom_tag.py"
        here = Path(__file__).parent
        src = here / "custom_tag.py"
        plugin.write_text(src.read_text())

        # copy config envs
        for src in (here / ".env", here / ".env"):
            dst = base / src.name
            dst.write_text(src.read_text())

        # set config root
        monkeypatch.setenv("GAMMA_CONFIG_ROOT", str(root))

        # update sys.path
        sys.path.insert(0, str(td))

        yield

        # cleanup sys.path
        sys.path.pop(0)

        # cleanup os.environ
        del os.environ["DUMMY_ENV"]


def test_custom_plugins(folder_fixture):
    import foo.custom_tag  # noqa
    from gamma.config import render_node, ScalarNode, Tag

    # check dispatch correctly added the custom renderer
    assert render_node[ScalarNode, Tag["!myenv"]] is not None

    # load config
    config = get_config()

    assert config["test"] == os.getenv("USER")
    assert config["test2"] == os.getenv("USER")
    assert os.getenv("DUMMY_ENV")



import pytest
from gamma.dispatch.poset import PODict, POSet


class PO:
    def __init__(self, val) -> None:
        self.val = val

    def __lt__(self, o) -> bool:
        if type(self.val) is not type(o.val):  # noqa
            return False
        return self.val < o.val

    def __le__(self, o) -> bool:
        if type(self.val) is not type(o.val):  # noqa
            return False
        return self.val <= o.val

    def __eq__(self, o: object) -> bool:
        if type(self.val) is not type(o.val):  # noqa
            return False
        return self.val == o.val

    def __repr__(self) -> str:
        return str(self.val)


def test_podict():

    pa = PO("a")
    pb = PO("b")
    p1 = PO(1)
    p2 = PO(2)
    assert pa < pb
    assert p1 < p2
    assert not (pa <= p1) and not (p1 <= pa)

    q = [(pb, 1), (p2, 1), (pa, 1), (p1, 1)]
    podict = PODict(q)
    assert list(podict.keys()) in (
        [pa, pb, p1, p2],
        [p1, p2, pa, pb],
        [p1, pa, p2, pb],
        [pa, p1, p2, pb],
        [pa, p1, pb, p2],
        [p1, pa, pb, p2],
    )

    del podict[pb]
    assert list(podict.keys()) in ([pa, p1, p2], [p1, p2, pa], [p1, pa, p2])

    podict[pa] = 100
    assert list(podict.keys()) in ([pa, p1, p2], [p1, p2, pa], [p1, pa, p2])
    assert podict[pa] == 100

    with pytest.raises(KeyError):
        podict[pb]

    podict[pb] = 200
    got = podict.pop(pb)
    assert got == 200
    assert list(podict.keys()) in ([pa, p1, p2], [p1, p2, pa], [p1, pa, p2])

    with pytest.raises(KeyError):
        podict.pop(pb)
    assert 1 == podict.pop(pb, 1)

    assert list(podict) == list(podict.keys())
    assert list(x[0] for x in podict.items()) == list(podict)
    assert list(x[1] for x in podict.items()) == list(podict.values())
    assert pa in podict

    podict.clear()
    assert len(podict) == 0

    poset = POSet([pb, p1, pa, p2])
    assert list(poset) in (
        [pa, pb, p1, p2],
        [p1, p2, pa, pb],
        [p1, pa, p2, pb],
        [pa, p1, p2, pb],
        [pa, p1, pb, p2],
        [p1, pa, pb, p2],
    )
    poset.remove(pa)
    poset.remove(pb)
    assert list(poset) == [p1, p2]

    poset.add(pa)
    assert list(poset) in ([pa, p1, p2], [p1, p2, pa], [p1, pa, p2])



import pytest
from gamma.dispatch import dispatch as base_dispatch
from gamma.dispatch.dispatchsystem import DispatchError

dispatch = base_dispatch(namespace="myns")
specialize = base_dispatch(namespace="myns", specialize=True)


def _create_foo_1():
    @dispatch
    def foo(a: int):
        return "int"

    return foo


def _create_foo_2():
    @dispatch
    def foo(a: str):
        return "str"

    return foo


def test_namespaced_and_special():

    # specialize should not work here
    with pytest.raises(DispatchError):

        @specialize
        def foo(a):
            return "obj"

    # create out of scope functions
    _create_foo_1()
    _create_foo_2()

    # create a third binding function
    @specialize
    def foo(a: float):
        return "float"

    assert foo(1) == "int"
    assert foo("a") == "str"
    assert foo(1.0) == "float"



from gamma.config.load import load_node
from gamma.config.rawnodes import as_node, get_item, get_values, is_equal


def test_get_item():
    # test simple
    left = load_node("foo: bar")
    right = load_node("foo: baz")

    assert is_equal(get_item(left, "foo"), "bar")
    assert is_equal(get_item(right, "foo"), "baz")


def test_get_values():
    # test simple
    node = load_node("[a, b, c]")

    items = get_values(node)
    expect = [as_node(x) for x in ["a", "b", "c"]]
    assert all([is_equal(a, b) for a, b in zip(items, expect)])



# let's load the gamma-config plugins module
import os

from gamma.config import ScalarNode, Tag, render_node
from gamma.dispatch import dispatch

# create a tag handler for !myenv
MyEnvTag = Tag["!myenv"]


@dispatch
def render_node(node: ScalarNode, tag: MyEnvTag, **ctx) -> str:
    """Simplyfied !env tag without default handling"""
    env_val = os.getenv(node.value)
    return env_val



import pytest


def test_default_ctx():
    from gamma.config import get_config
    from gamma.config.render_context import get_render_context

    get_config()  # requires initialization
    ctx = get_render_context()

    assert "include_folders" in ctx["c"]  # from 00-meta.yaml
    assert ctx["env"]["USER"]


@pytest.fixture
def custom_ctx():
    from gamma.config import context_providers, ContextVar
    from gamma.config.cache import cache
    import uuid

    var_foo = ContextVar("foo", "bar")
    var_func = ContextVar("func", function=lambda: "zit")
    var_cache = ContextVar("cache", function=lambda: str(uuid.uuid4()), cacheable=True)
    var_rnd = ContextVar("rnd", function=lambda: str(uuid.uuid4()))
    prov = [var_foo, var_func, var_cache, var_rnd]

    var_foo2 = ContextVar("foo2", "bar")
    prov2 = lambda: [var_foo2]  # noqa

    context_providers.append(prov)
    context_providers.append(prov2)

    yield

    context_providers.pop()
    context_providers.pop()
    cache.clear()


def test_custom_ctx(custom_ctx):
    from gamma.config.render_context import get_render_context

    ctx = get_render_context()

    assert ctx["foo"] == "bar"
    assert ctx["func"] == "zit"
    old_cache = ctx["cache"]
    old_rnd = ctx["rnd"]

    ctx = get_render_context()
    new_cache = ctx["cache"]
    new_rnd = ctx["rnd"]

    assert new_cache == old_cache
    assert new_rnd != old_rnd

    assert ctx["foo2"] == "bar"



import pytest
from gamma.config.load import load_node
from gamma.config.render import RenderDispatchError, render_node
from gamma.config.tags import Tag
from gamma.dispatch import dispatch
from ruamel.yaml.nodes import Node


def test_scalar_str():
    node_a = load_node("foo: bar")
    got = render_node(node_a)
    want = {"foo": "bar"}
    assert got == want


def test_scalar_seq():
    node_a = load_node("foo: [1, 2, 3]")
    got = render_node(node_a)
    want = {"foo": [1, 2, 3]}
    assert got == want


def test_scalar_int():
    assert render_node(load_node("foo: 99"))["foo"] == 99
    assert render_node(load_node("foo: 0o777"))["foo"] == 0o777
    assert render_node(load_node("foo: 0xBEEF"))["foo"] == 0xBEEF


def test_scalar_float():
    node_a = load_node("foo: 1.0")
    got = render_node(node_a)
    want = {"foo": 1.0}
    assert got == want


def test_scalar_null():
    assert render_node(load_node("foo: null"))["foo"] is None
    assert render_node(load_node("foo: Null"))["foo"] is None
    assert render_node(load_node("foo: NULL"))["foo"] is None
    assert render_node(load_node("foo: "))["foo"] is None
    assert render_node(load_node("foo: ~"))["foo"] is None


def test_scalar_bool():
    node_a = load_node("foo: True")
    node_b = load_node("foo: true")
    node_c = load_node("foo: TRUE")

    got_a = render_node(node_a)
    got_b = render_node(node_b)
    got_c = render_node(node_c)
    want = {"foo": True}
    assert got_a == want
    assert got_b == want
    assert got_c == want

    # YAML 1.2 requires only "true" and variants to autocast to bool
    assert render_node(load_node("foo: yes")) == {"foo": "yes"}
    assert render_node(load_node("foo: !!bool yes")) == {"foo": True}
    assert render_node(load_node("foo: !!bool no")) == {"foo": False}


def test_scalar_datetime():
    for v in ("2020-01-30", "2020-01-30T10:11:12", "2020-01-30T10:11:12Z"):
        node = load_node(f"foo: {v}")
        got = render_node(node)
        want = {"foo": v}  # should not render as date objects
        assert got == want


def test_scalar_nan():
    node = load_node("foo: .nan")
    got = render_node(node)["foo"]
    assert got != got  # NaN's are weird


def test_scalar_inf():
    node = load_node("foo: .inf")
    got = render_node(node)["foo"]
    assert got == float("inf")


def test_render_uri():
    from gamma.config import render_node

    src = """
    a: !foo     1
    b: !bar:b   2
    c: !bar:c   3
    """
    node = load_node(src)

    with pytest.raises(RenderDispatchError):
        render_node(node)

    Foo = Tag["!foo"]
    Bar = Tag["!bar"]
    BarC = Tag["!bar:c"]

    try:

        @dispatch
        def render_node(node: Node, tag: Foo, **ctx):
            return f"foo-{node.value}"

        @dispatch
        def render_node(node: Node, tag: Bar, **ctx):
            assert ctx["path"]
            return f"bar-{node.value}"

        @dispatch
        def render_node(node: Node, tag: BarC, **ctx):
            assert ctx.get("path") is None
            return f"bar-c-{node.value}"

        d = render_node(node)
        assert d["a"] == "foo-1"
        assert d["b"] == "bar-2"
        assert d["c"] == "bar-c-3"

    finally:
        del render_node[Node, Foo]
        del render_node[Node, Bar]
        del render_node[Node, BarC]



import time

import pytest
from gamma.config.confignode import (
    ConfigNode,
    RootConfig,
    config_context,
    push_entry,
    push_folder,
)
from gamma.config.globalconfig import get_config, set_config
from gamma.config.load import load_node

SIMPLE = """
foo:
    bar: 1
    zoo: [1, 2, 3]
    sub:
        bar: 10
    zit:
"""


def test_simple_config():

    node = load_node(SIMPLE)
    cfg = ConfigNode(node)
    assert isinstance(cfg["foo"], ConfigNode)
    assert cfg["foo"]["bar"] == 1
    assert cfg["foo"]["zoo"] == [1, 2, 3]
    assert cfg["foo"]["sub"]["bar"] == 10
    assert cfg["foo"]["zit"] is None


def test_attribute_dot_access():
    META = SIMPLE + "\n__enable_dot_access__: True"
    cfg = RootConfig("00-meta.yaml", META)
    assert cfg.foo.bar == 1
    assert cfg.foo.zoo == [1, 2, 3]
    assert cfg.foo.sub.bar == 10

    with pytest.raises(AttributeError):
        assert bool(cfg.foo.notexist) == False  # noqa


def test_guards():
    with pytest.raises(Exception, match="entry"):
        RootConfig("foo: 1")

    cfg = RootConfig()
    push_entry(cfg, "abc", "foo: 1")

    assert cfg["foo"] == 1

    with pytest.raises(Exception, match="duplicated"):
        push_entry(cfg, "abc", "foo: 1")

    with pytest.raises(Exception, match="entry"):
        push_entry(cfg, "~abc", "foo: 1")


def test_config_context():
    cfg = RootConfig()
    push_entry(cfg, "abc", "foo: 1")

    assert cfg["foo"] == 1

    with config_context(cfg, "foo: 2"):
        assert cfg["foo"] == 2

    assert cfg["foo"] == 1

    set_config(cfg)
    with config_context("foo: 3"):
        get_config()["foo"] == 3

    assert cfg["foo"] == 1

    with pytest.raises(Exception):
        with config_context("foo: 3"):
            raise ValueError("Random exception")

    assert cfg["foo"] == 1

    # stress testing
    for _ in range(10):
        with pytest.raises(Exception):
            with config_context("foo: 3"):
                raise ValueError("Random exception")
        time.sleep(0.05)
        assert cfg["foo"] == 1


def test_push_folder():
    from pathlib import Path

    folder = Path(__file__).parent / "config_folder"

    cfg = RootConfig()
    push_entry(cfg, "10-base.yaml", SIMPLE)

    assert cfg["foo"]["bar"] != 1000
    assert cfg["foo"]["zit"] != 1000

    push_folder(cfg, folder)

    assert cfg["foo"]["bar"] == 1000
    assert cfg["foo"]["zit"] == 1000



import os
from pathlib import Path
from typing import NamedTuple

from gamma.config import RootConfig, ScalarNode, render_node, to_dict
from gamma.config.builtin_tags import RefTag


def test_ref():
    # simple access
    root = RootConfig("dummy", {"foo": {"bar": 100}})
    node = ScalarNode("!ref", "foo.bar")
    val = render_node(node, RefTag(), config=root)
    assert val == 100

    # quoted access
    root = RootConfig("dummy", {"hello.world": {"bar": 100}})
    node = ScalarNode("!ref", "'hello.world'.bar")
    val = render_node(node, RefTag(), config=root)
    assert val == 100


def test_ref_sibling():
    src = """
    a: foo
    b:
      ref_a: !ref a
      sib: !ref b.ref_a
    """

    cfg = RootConfig("dummy", src)
    assert cfg["b"]["ref_a"] == "foo"
    assert cfg["a"] == "foo"
    assert cfg["b"]["sib"] == "foo"


def test_ref_to_dict():
    src = """
    a: foo
    c:
      sub: bar
    b:
      ref_a: !ref a
      sib: !ref b.ref_a
      ref_c: !ref c
    """

    cfg = RootConfig("dummy", src)
    d = to_dict(cfg, recursive=True)
    assert d["b"]["ref_a"] == "foo"
    assert isinstance(d["b"]["ref_c"], dict)


class MyObj(NamedTuple):
    val: dict


class MyObj2(NamedTuple):
    a: int
    b: int


def make_double(val):
    return [2 * x for x in val]


def make_list(val):
    return [val]


def hello():
    return "world"


def test_py_tag():
    mod = __name__
    src = f"""
    foo: !py:{mod}:MyObj
        a: !py:{mod}:make_list 1
        b: !py:{mod}:make_double
            - 2
    hello: !py:{mod}:hello
    """

    cfg = RootConfig("dummy", src)
    foo = cfg["foo"]
    assert isinstance(foo, MyObj)
    assert isinstance(foo.val, dict)
    assert foo.val["a"] == [1]
    assert foo.val["b"] == [4]
    assert cfg["hello"] == "world"


def test_obj_tag():
    mod = __name__
    src = f"""
    obj_default_module: {mod}

    foo: !obj:MyObj2
        a: 1
        b: 2
    """

    foo = RootConfig("dummy", src)["foo"]
    assert isinstance(foo, MyObj2)
    assert foo.a == 1
    assert foo.b == 2


def test_path_tag():
    # construct paths for testing non-OS-specific:
    test_path = os.path.join("testdir", "testfile")
    test_path_fragment = os.path.join(os.pardir, test_path)

    src = f"""
    foo: !path ..{test_path_fragment}
    """

    foo = RootConfig("dummy", src)["foo"]

    assert str(foo).endswith(test_path_fragment)
    assert Path(foo).is_absolute()



import typing
from typing import Any, Tuple, Type, Union

from gamma.dispatch.typesystem import Sig, SigSet, Vararg, is_more_specific, issubtype


class SuperFoo:
    pass


class Foo(SuperFoo):
    pass


class SuperBar:
    pass


class Bar(SuperBar):
    pass


def test_partial_order_unary():
    foo = Tuple[Foo]
    base = Tuple[SuperFoo]
    other = Tuple[SuperBar]
    obj = Tuple[object]

    assert issubtype(foo, obj)
    assert issubtype(foo, base)
    assert is_more_specific(foo, base)
    assert issubtype(foo, foo)
    assert not (issubtype(base, foo))
    assert not (is_more_specific(foo, foo))
    assert not (issubtype(foo, other))
    assert not (is_more_specific(foo, other))


def test_sort():
    a = Sig(object, object)
    b = Sig(object)
    c = Sig(SuperFoo, object)

    assert is_more_specific(c, a)
    table = SigSet([a, b, c])
    assert list(table) in ([c, b, a], [c, a, b], [b, c, a])


def test_strict_subset():
    a = Sig(SuperFoo, object)
    b = Sig(object, object)
    assert is_more_specific(a, b)
    assert not (is_more_specific(a, a))
    assert not (is_more_specific(b, b))


def test_partial_order_nary():
    # dispatch is check if issubtype(call tuple, target tuple)
    call = Sig(Foo, Foo)
    target = Sig(object, object)
    assert issubtype(call, target)

    call = Sig(Foo, object)
    target = Sig(object, object)
    assert issubtype(call, target)

    call = Sig(Foo, SuperFoo)
    target = Sig(SuperFoo, SuperFoo)
    assert issubtype(call, target)

    call = Sig(Foo, Foo)
    target = Sig(SuperFoo, SuperFoo)
    assert issubtype(call, target)

    call = Sig(Foo, Foo)
    target = Sig(Foo, Foo)
    assert issubtype(call, target)

    # check arity
    call = Sig(
        Foo,
    )
    target = Sig(Foo, Foo)
    assert not (issubtype(call, target))


def test_issubtype():

    # strict check
    assert not (issubtype(Foo, Bar))
    assert not (issubtype(Bar, Foo))
    assert issubtype(Foo, Foo)
    assert issubtype(Foo, SuperFoo)
    assert not issubtype(Bar, SuperFoo)
    assert issubtype(Foo, Union[SuperFoo, SuperBar])
    assert issubtype(Bar, Union[SuperFoo, SuperBar])
    assert issubtype(Bar, Union[Foo, Bar])


def test_issubtype_special():
    assert issubtype(Type[int], type)
    assert issubtype(Type[int], Type)
    assert issubtype(int, Any)


def test_equality():
    a = Sig(float, str)
    b = Sig(
        float,
    )
    assert not a == b


def test_varargs():

    target = Sig(int, int, Vararg)
    assert issubtype(Sig(int, int), target)

    call = Sig(int, int, int)
    assert issubtype(call, target)


def test_poset():
    class Tag:
        pass

    class ATag(Tag):
        pass

    class Node:
        pass

    class ScalarNode(Node):
        pass

    s1 = Sig(Node, Tag)
    s2 = Sig(
        Node,
    )
    s3 = Sig(ScalarNode, ATag)
    assert issubtype(s3, s1)
    assert is_more_specific(s3, s1)
    assert not is_more_specific(s2, s1)


def test_type_lit():
    assert issubtype(typing.Type[str], typing.Type[object])
    assert issubtype(typing.Type[str], type)
    assert not issubtype(type, typing.Type[str])
    assert issubtype(typing.Type[Foo], typing.Type[SuperFoo])
    assert not issubtype(typing.Type[Foo], typing.Type[Bar])



from typing import List

from gamma.config import RootConfig
from gamma.config.pydantic import ConfigStruct


class Dataset(ConfigStruct):
    """Base class with common properties."""

    path: str
    compression: str


class CsvDataset(Dataset):
    kind = "csv"
    separator: str


class ParquetDataset(Dataset):
    kind = "parquet"
    columns: List[str]


# note we're using a function, not a class
def Datasets(**entries):
    return {key: Dataset.parse_obj(obj) for key, obj in entries.items()}


def test_auto2():
    from tests.gamma.config.pydantic.test_doc4 import CsvDataset, ParquetDataset

    src = """
obj_default_module: tests.gamma.config.pydantic.test_doc4

datasets: !obj:Datasets
  foo:
    kind: csv
    path: data/foo.csv.gz
    compression: gzip
    separator: ";"

  bar:
    kind: parquet
    path: data/bar.parquet
    compression: snappy
    columns: [col_x, col_y]
    """

    config = RootConfig("dummy", src)
    get_config = lambda: config  # noqa

    foo = get_config()["datasets"]["foo"]
    bar = get_config()["datasets"]["bar"]
    assert isinstance(foo, CsvDataset)
    assert isinstance(bar, ParquetDataset)



import os
import tempfile
import time
from pathlib import Path

import pytest


@pytest.fixture
def test_cwd():
    with tempfile.TemporaryDirectory() as td:
        cwd = os.getcwd()
        os.chdir(td)
        yield td
        os.chdir(cwd)


def test_scaffold(test_cwd):
    from gamma.config.scaffold import scaffold

    # test cwd
    scaffold(target=None, force=False)
    meta = Path(test_cwd) / "config/00-meta.yaml"
    assert meta.exists()
    expect = meta.stat().st_ctime_ns

    # assert no overwrite
    time.sleep(0.1)
    scaffold(target=None, force=False)
    got = meta.stat().st_ctime_ns
    assert expect == got

    # test target folder
    with tempfile.TemporaryDirectory() as td:
        scaffold(target=td, force=False)
        assert (Path(td) / "config/00-meta.yaml").exists()


def test_missing_meta_warning(test_cwd):
    from gamma.config.scaffold import scaffold
    from gamma.config import get_config
    from gamma.config.findconfig import MissingMetaConfigFile

    # test cwd
    scaffold(target=None, force=False)
    meta = Path(test_cwd) / "config/00-meta.yaml"
    assert meta.exists()

    # remove meta for testing
    meta.unlink()

    # load config
    with pytest.raises(MissingMetaConfigFile):
        _ = get_config()
